# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.hub.iothub_http_runtime_manager import IoTHubHttpRuntimeManager
from azure.iot.hub.auth import ConnectionStringAuthentication
from azure.iot.hub.protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs

"""---Constants---"""

fake_hostname = "beauxbatons.academy-net"
fake_device_id = "MyPensieve"
fake_shared_access_key_name = "alohomora"
fake_shared_access_key = "Zm9vYmFy"
fake_lock_token = "fake_lock_token"


"""----Shared fixtures----"""


@pytest.fixture(scope="function", autouse=True)
def mock_http_runtime_operations(mocker):
    mock_http_runtime_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.CloudToDeviceMessagesOperations"
    )
    return mock_http_runtime_operations_init.return_value


@pytest.fixture(scope="function")
def iothub_http_runtime_manager():
    connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
        hostname=fake_hostname,
        device_id=fake_device_id,
        skn=fake_shared_access_key_name,
        sk=fake_shared_access_key,
    )
    iothub_http_runtime_manager = IoTHubHttpRuntimeManager.from_connection_string(connection_string)
    return iothub_http_runtime_manager


@pytest.mark.describe("IoTHubHttpRuntimeManager - .from_connection_string()")
class TestFromConnectionString:
    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
                    hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
                ),
                id="connection string without SharedAccessKeyName",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of ConnectionStringAuthentication and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_connection_string_auth(self, connection_string):
        client = IoTHubHttpRuntimeManager.from_connection_string(
            connection_string=connection_string
        )

        assert repr(client.auth) == connection_string
        assert client.protocol.config.base_url == "https://" + client.auth["HostName"]
        assert client.protocol.config.credentials == client.auth

    @pytest.mark.it("Sets the auth and protocol attributes")
    def test_instantiates_auth_and_protocol_attributes(self, iothub_http_runtime_manager):
        assert isinstance(iothub_http_runtime_manager.auth, ConnectionStringAuthentication)
        assert isinstance(iothub_http_runtime_manager.protocol, IotHubGatewayServiceAPIs)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with an empty connection string"
    )
    def test_instantiates_with_empty_connection_string(self):
        with pytest.raises(ValueError):
            IoTHubHttpRuntimeManager.from_connection_string("")

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without HostName"
    )
    def test_instantiates_with_connection_string_no_host_name(self):
        connection_string = (
            "DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                device_id=fake_device_id, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        with pytest.raises(ValueError):
            IoTHubHttpRuntimeManager.from_connection_string(connection_string)

    @pytest.mark.it("Instantiates with an connection string without DeviceId")
    def test_instantiates_with_connection_string_no_device_id(self):
        connection_string = (
            "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                hostname=fake_hostname, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        obj = IoTHubHttpRuntimeManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubHttpRuntimeManager)

    @pytest.mark.it("Instantiates with an connection string without SharedAccessKeyName")
    def test_instantiates_with_connection_string_no_shared_access_key_name(self):
        connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
            hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
        )
        obj = IoTHubHttpRuntimeManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubHttpRuntimeManager)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without SharedAccessKey"
    )
    def test_instantiates_with_connection_string_no_shared_access_key(self):
        connection_string = (
            "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn}".format(
                hostname=fake_hostname, device_id=fake_device_id, skn=fake_shared_access_key_name
            )
        )
        with pytest.raises(ValueError):
            IoTHubHttpRuntimeManager.from_connection_string(connection_string)


@pytest.mark.describe("IoTHubHttpRuntimeManager - .from_token_credential()")
class TestFromTokenCredential:
    @pytest.mark.it(
        "Creates an instance of AzureIdentityCredentialAdapter and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_token_credential_auth(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()

        client = IoTHubHttpRuntimeManager.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential
        )

        assert client.auth._policy._credential == mock_azure_identity_TokenCredential
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert client.protocol.config.credentials == client.auth


@pytest.mark.describe("IoTHubHttpRuntimeManager - .receive_feedback_notification()")
class TestReceiveFeedbackNotification(object):
    @pytest.mark.it("Uses protocol layer HTTP runtime to receive feedback notifications")
    def test_receive_feedback_notification(
        self, mocker, mock_http_runtime_operations, iothub_http_runtime_manager
    ):
        ret_val = iothub_http_runtime_manager.receive_feedback_notification()
        assert mock_http_runtime_operations.receive_feedback_notification.call_count == 1
        assert mock_http_runtime_operations.receive_feedback_notification.call_args == mocker.call()
        assert ret_val == mock_http_runtime_operations.receive_feedback_notification()


@pytest.mark.describe("IoTHubHttpRuntimeManager - .complete_feedback_notification()")
class TestCompleteFeedbackNotification(object):
    @pytest.mark.it("Uses protocol layer HTTP runtime to complete feedback notifications")
    def test_complete_feedback_notification(
        self, mocker, mock_http_runtime_operations, iothub_http_runtime_manager
    ):
        ret_val = iothub_http_runtime_manager.complete_feedback_notification(fake_lock_token)
        assert mock_http_runtime_operations.complete_feedback_notification.call_count == 1
        assert mock_http_runtime_operations.complete_feedback_notification.call_args == mocker.call(
            fake_lock_token
        )
        assert ret_val == mock_http_runtime_operations.complete_feedback_notification()


@pytest.mark.describe("IoTHubHttpRuntimeManager - .abandon_feedback_notification()")
class TestAbandonFeedbackNotification(object):
    @pytest.mark.it("Uses protocol layer HTTP runtime to abandon feedback notifications")
    def test_abandon_feedback_notification(
        self, mocker, mock_http_runtime_operations, iothub_http_runtime_manager
    ):
        ret_val = iothub_http_runtime_manager.abandon_feedback_notification(fake_lock_token)
        assert mock_http_runtime_operations.abandon_feedback_notification.call_count == 1
        assert mock_http_runtime_operations.abandon_feedback_notification.call_args == mocker.call(
            fake_lock_token
        )
        assert ret_val == mock_http_runtime_operations.abandon_feedback_notification()



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.hub.protocol.models import AuthenticationMechanism, DeviceCapabilities
from azure.iot.hub.iothub_registry_manager import IoTHubRegistryManager
from azure.iot.hub import iothub_amqp_client
from azure.iot.hub.protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs
from uamqp import TransportType

"""---Constants---"""

fake_shared_access_key = "Zm9vYmFy"
fake_shared_access_key_name = "alohomora"

fake_primary_key = "petrificus"
fake_secondary_key = "totalus"
fake_primary_thumbprint = "HELFKCPOXAIR9PVNOA3"
fake_secondary_thumbprint = "RGSHARLU4VYYFENINUF"
fake_hostname = "beauxbatons.academy-net"
fake_device_id = "MyPensieve"
fake_module_id = "Divination"
fake_managed_by = "Hogwarts"
fake_etag = "taggedbymisnitryofmagic"
fake_status = "flying"
fake_status_reason = "some_reason"
fake_configuration_id = "fake_configuration"
fake_configuration = "fake_config"
fake_max_count = 42
fake_configuration_queries = "fake_configuration_queries"
fake_devices = "fake_devices"
fake_query_specification = "fake_query_specification"
fake_configuration_content = "fake_configuration_content"
fake_job_id = "fake_job_id"
fake_start_time = "fake_start_time"
fake_end_time = "fake_end_time"
fake_job_type = "fake_job_type"
fake_job_request = "fake_job_request"
fake_job_status = "fake_status"
fake_job_properties = "fake_job_properties"
fake_device_twin = "fake_device_twin"
fake_module_twin = "fake_module_twin"
fake_direct_method_request_with_payload_none = "fake_direct_method_request"
fake_properties = {"fake_key": "fake_value"}
fake_device_scope = "fake_device_scope"
fake_parent_scopes = ["fake_parent_scopes"]


class Fake_direct_method_request_with_payload:
    payload = ""


fake_direct_method_request_with_payload = Fake_direct_method_request_with_payload()


class Fake_direct_method_request_with_payload_none:
    payload = None


fake_direct_method_request_with_payload_none = Fake_direct_method_request_with_payload_none()
fake_message_to_send = "fake_message_to_send"

"""----Shared fixtures----"""


@pytest.fixture(scope="function", autouse=True)
def mock_devices_operations(mocker):
    mock_devices_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.DevicesOperations"
    )
    return mock_devices_operations_init.return_value


@pytest.fixture(scope="function", autouse=True)
def mock_modules_operations(mocker):
    mock_modules_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.ModulesOperations"
    )
    return mock_modules_operations_init.return_value


@pytest.fixture(scope="function", autouse=True)
def mock_statistics_operations(mocker):
    mock_statistics_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.StatisticsOperations"
    )
    return mock_statistics_operations_init.return_value


@pytest.fixture(scope="function", autouse=True)
def mock_bulk_registry_operations(mocker):
    mock_bulk_registry_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.BulkRegistryOperations"
    )
    return mock_bulk_registry_operations_init.return_value


@pytest.fixture(scope="function", autouse=True)
def mock_query_operations(mocker):
    mock_query_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.QueryOperations"
    )
    return mock_query_operations_init.return_value


@pytest.fixture(scope="function")
def iothub_registry_manager():
    connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
        hostname=fake_hostname,
        device_id=fake_device_id,
        skn=fake_shared_access_key_name,
        sk=fake_shared_access_key,
    )
    iothub_registry_manager = IoTHubRegistryManager.from_connection_string(connection_string)
    return iothub_registry_manager


@pytest.fixture(scope="function")
def mock_device_constructor(mocker):
    return mocker.patch("azure.iot.hub.iothub_registry_manager.Device")


@pytest.fixture(scope="function")
def mock_module_constructor(mocker):
    return mocker.patch("azure.iot.hub.iothub_registry_manager.Module")


@pytest.fixture(scope="function")
def mock_uamqp_send_message_to_device(mocker):
    mock_uamqp_send = mocker.patch.object(
        iothub_amqp_client.IoTHubAmqpClientSharedAccessKeyAuth,
        "send_message_to_device",
    )
    return mock_uamqp_send


@pytest.mark.describe("IoTHubRegistryManager - .from_connection_string()")
class TestFromConnectionString:
    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of IotHubGatewayServiceAPIs and IoTHubAmqpClientSharedAccessKeyAuth with the correct arguments"
    )
    def test_connection_string_auth(self, mocker, connection_string):
        amqp_client_init_mock = mocker.patch.object(
            iothub_amqp_client, "IoTHubAmqpClientSharedAccessKeyAuth"
        )

        client = IoTHubRegistryManager.from_connection_string(connection_string=connection_string)

        assert repr(client.protocol.config.credentials) == connection_string
        assert (
            client.protocol.config.base_url
            == "https://" + client.protocol.config.credentials["HostName"]
        )
        assert amqp_client_init_mock.call_args == mocker.call(
            client.protocol.config.credentials["HostName"],
            client.protocol.config.credentials["SharedAccessKeyName"],
            client.protocol.config.credentials["SharedAccessKey"],
            TransportType.Amqp,
        )

    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of IotHubGatewayServiceAPIs and IoTHubAmqpClientSharedAccessKeyAuth with the correct arguments and using AMQP over Websocket"
    )
    def test_connection_string_auth_amqp_over_websocket(self, mocker, connection_string):
        amqp_client_init_mock = mocker.patch.object(
            iothub_amqp_client, "IoTHubAmqpClientSharedAccessKeyAuth"
        )

        client = IoTHubRegistryManager.from_connection_string(connection_string=connection_string, transport_type=TransportType.AmqpOverWebsocket)

        assert repr(client.protocol.config.credentials) == connection_string
        assert (
            client.protocol.config.base_url
            == "https://" + client.protocol.config.credentials["HostName"]
        )
        assert amqp_client_init_mock.call_args == mocker.call(
            client.protocol.config.credentials["HostName"],
            client.protocol.config.credentials["SharedAccessKeyName"],
            client.protocol.config.credentials["SharedAccessKey"],
            TransportType.AmqpOverWebsocket,
        )

    @pytest.mark.it("Sets the protocol attribute")
    def test_instantiates_auth_and_protocol_attributes(self, iothub_registry_manager):
        assert isinstance(iothub_registry_manager.protocol, IotHubGatewayServiceAPIs)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with an empty connection string"
    )
    def test_instantiates_with_empty_connection_string(self):
        with pytest.raises(ValueError):
            IoTHubRegistryManager.from_connection_string("")

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without HostName"
    )
    def test_instantiates_with_connection_string_no_host_name(self):
        connection_string = (
            "DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                device_id=fake_device_id, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        with pytest.raises(ValueError):
            IoTHubRegistryManager.from_connection_string(connection_string)

    @pytest.mark.it("Instantiates with an connection string without DeviceId")
    def test_instantiates_with_connection_string_no_device_id(self):
        connection_string = (
            "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                hostname=fake_hostname, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        obj = IoTHubRegistryManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubRegistryManager)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without SharedAccessKey"
    )
    def test_instantiates_with_connection_string_no_shared_access_key(self):
        connection_string = (
            "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn}".format(
                hostname=fake_hostname, device_id=fake_device_id, skn=fake_shared_access_key_name
            )
        )
        with pytest.raises(ValueError):
            IoTHubRegistryManager.from_connection_string(connection_string)


@pytest.mark.describe("IoTHubRegistryManager - .from_token_credential()")
class TestFromTokenCredential:
    @pytest.mark.it(
        "Creates an instance of IotHubGatewayServiceAPIs and IoTHubAmqpClientTokenAuth with the correct arguments"
    )
    def test_token_credential_auth(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()
        amqp_client_init_mock = mocker.patch.object(iothub_amqp_client, "IoTHubAmqpClientTokenAuth")

        client = IoTHubRegistryManager.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential
        )

        assert (
            client.protocol.config.credentials._policy._credential
            == mock_azure_identity_TokenCredential
        )
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert amqp_client_init_mock.call_args == mocker.call(
            fake_hostname, mock_azure_identity_TokenCredential, transport_type=TransportType.Amqp
        )
    def test_token_credential_auth_with_amqp_over_websocket(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()
        amqp_client_init_mock = mocker.patch.object(iothub_amqp_client, "IoTHubAmqpClientTokenAuth")

        client = IoTHubRegistryManager.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential, TransportType.AmqpOverWebsocket
        )

        assert (
            client.protocol.config.credentials._policy._credential
            == mock_azure_identity_TokenCredential
        )
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert amqp_client_init_mock.call_args == mocker.call(
            fake_hostname, mock_azure_identity_TokenCredential, transport_type=TransportType.AmqpOverWebsocket
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_device_with_sas()")
class TestCreateDeviceWithSymmetricKey(object):

    testdata = [
        (fake_primary_key, None),
        (None, fake_secondary_key),
        (fake_primary_key, fake_secondary_key),
    ]

    @pytest.mark.it("Initializes device with the provided parameters and sas auth")
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_initializes_device_with_kwargs_for_sas(
        self, iothub_registry_manager, mock_device_constructor, primary_key, secondary_key
    ):
        iothub_registry_manager.create_device_with_sas(
            device_id=fake_device_id,
            status=fake_status,
            primary_key=primary_key,
            secondary_key=secondary_key,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1

        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "sas"
        assert auth_mechanism.x509_thumbprint is None
        sym_key = auth_mechanism.symmetric_key
        assert sym_key.primary_key == primary_key
        assert sym_key.secondary_key == secondary_key

    @pytest.mark.it(
        "Calls method from service operations with device id and previously constructed device"
    )
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_calls_create_or_update_identity_for_sas(
        self,
        mock_device_constructor,
        mock_devices_operations,
        iothub_registry_manager,
        primary_key,
        secondary_key,
    ):
        iothub_registry_manager.create_device_with_sas(
            device_id=fake_device_id,
            status=fake_status,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_device_with_x509()")
class TestCreateDeviceWithX509(object):

    testdata = [
        (fake_primary_thumbprint, None),
        (None, fake_secondary_thumbprint),
        (fake_primary_thumbprint, fake_secondary_thumbprint),
    ]

    @pytest.mark.it("Initializes device with the provided parameters and X509 auth")
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_initializes_device_with_kwargs_for_x509(
        self,
        iothub_registry_manager,
        mock_device_constructor,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.create_device_with_x509(
            device_id=fake_device_id,
            status=fake_status,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1
        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "selfSigned"
        assert auth_mechanism.symmetric_key is None
        x509_thumbprint = auth_mechanism.x509_thumbprint
        assert x509_thumbprint.primary_thumbprint == primary_thumbprint
        assert x509_thumbprint.secondary_thumbprint == secondary_thumbprint

    @pytest.mark.it(
        "Calls method from service operations with device id and previously constructed device"
    )
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_calls_create_or_update_identity_for_x509(
        self,
        mock_device_constructor,
        mock_devices_operations,
        iothub_registry_manager,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.create_device_with_x509(
            device_id=fake_device_id,
            status=fake_status,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_device_with_certificate_authority()")
class TestCreateDeviceWithCA(object):
    @pytest.mark.it("Initializes device with the provided parameters and ca auth")
    def test_initializes_device_with_kwargs_for_certificate_authority(
        self, mock_device_constructor, iothub_registry_manager
    ):
        iothub_registry_manager.create_device_with_certificate_authority(
            device_id=fake_device_id,
            status=fake_status,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1
        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "certificateAuthority"
        assert auth_mechanism.x509_thumbprint is None
        assert auth_mechanism.symmetric_key is None

    @pytest.mark.it(
        "Calls method from service operations with device id and previously constructed device"
    )
    def test_calls_create_or_update_identity_for_certificate_authority(
        self, mock_device_constructor, mock_devices_operations, iothub_registry_manager
    ):
        iothub_registry_manager.create_device_with_certificate_authority(
            device_id=fake_device_id, status=fake_status
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_device_with_sas()")
class TestUpdateDeviceWithSymmetricKey(object):

    testdata = [
        (fake_primary_key, None),
        (None, fake_secondary_key),
        (fake_primary_key, fake_secondary_key),
    ]

    @pytest.mark.it("Initializes device with the provided parameters and sas auth")
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_initializes_device_with_kwargs_for_sas(
        self, iothub_registry_manager, mock_device_constructor, primary_key, secondary_key
    ):
        iothub_registry_manager.update_device_with_sas(
            device_id=fake_device_id,
            status=fake_status,
            etag=fake_etag,
            primary_key=primary_key,
            secondary_key=secondary_key,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1

        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "sas"
        assert auth_mechanism.x509_thumbprint is None
        sym_key = auth_mechanism.symmetric_key
        assert sym_key.primary_key == primary_key
        assert sym_key.secondary_key == secondary_key

    @pytest.mark.it(
        "Calls method from service operations with device id, etag, and previously constructed device"
    )
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_calls_create_or_update_identity_for_sas(
        self,
        mock_device_constructor,
        mock_devices_operations,
        iothub_registry_manager,
        primary_key,
        secondary_key,
    ):
        iothub_registry_manager.update_device_with_sas(
            device_id=fake_device_id,
            status=fake_status,
            etag=fake_etag,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][2]
            == '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_device_with_x509()")
class TestUpdateDeviceWithX509(object):

    testdata = [
        (fake_primary_thumbprint, None),
        (None, fake_secondary_thumbprint),
        (fake_primary_thumbprint, fake_secondary_thumbprint),
    ]

    @pytest.mark.it("Initializes device with the provided parameters and X509 auth")
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_initializes_device_with_kwargs_for_x509(
        self,
        iothub_registry_manager,
        mock_device_constructor,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.update_device_with_x509(
            device_id=fake_device_id,
            status=fake_status,
            etag=fake_etag,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1
        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "selfSigned"
        assert auth_mechanism.symmetric_key is None
        x509_thumbprint = auth_mechanism.x509_thumbprint
        assert x509_thumbprint.primary_thumbprint == primary_thumbprint
        assert x509_thumbprint.secondary_thumbprint == secondary_thumbprint

    @pytest.mark.it(
        "Calls method from service operations with device id, etag, and previously constructed device"
    )
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_calls_create_or_update_identity_for_x509(
        self,
        mock_device_constructor,
        mock_devices_operations,
        iothub_registry_manager,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.update_device_with_x509(
            device_id=fake_device_id,
            status=fake_status,
            etag=fake_etag,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][2]
            == '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_device_with_certificate_authority()")
class TestUpdateDeviceWithCA(object):
    @pytest.mark.it("Initializes device with the provided parameters and ca auth")
    def test_initializes_device_with_kwargs_for_certificate_authority(
        self, mock_device_constructor, iothub_registry_manager
    ):
        iothub_registry_manager.update_device_with_certificate_authority(
            device_id=fake_device_id,
            status=fake_status,
            etag=fake_etag,
            iot_edge=False,
            status_reason=fake_status_reason,
            device_scope=fake_device_scope,
            parent_scopes=fake_parent_scopes,
        )

        assert mock_device_constructor.call_count == 1
        assert mock_device_constructor.call_args[1]["device_id"] == fake_device_id
        assert mock_device_constructor.call_args[1]["status"] == fake_status
        assert mock_device_constructor.call_args[1]["status_reason"] == fake_status_reason
        assert mock_device_constructor.call_args[1]["device_scope"] == fake_device_scope
        assert mock_device_constructor.call_args[1]["parent_scopes"] == fake_parent_scopes
        assert isinstance(mock_device_constructor.call_args[1]["capabilities"], DeviceCapabilities)
        device_capabilities = mock_device_constructor.call_args[1]["capabilities"]
        assert device_capabilities.iot_edge is False
        assert isinstance(
            mock_device_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_device_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "certificateAuthority"
        assert auth_mechanism.x509_thumbprint is None
        assert auth_mechanism.symmetric_key is None

    @pytest.mark.it(
        "Calls method from service operations with device id, etag, and previously constructed device"
    )
    def test_calls_create_or_update_identity_for_certificate_authority(
        self, mock_device_constructor, mock_devices_operations, iothub_registry_manager
    ):
        iothub_registry_manager.update_device_with_certificate_authority(
            device_id=fake_device_id, status=fake_status, etag=fake_etag
        )

        assert mock_devices_operations.create_or_update_identity.call_count == 1
        assert mock_devices_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][1]
            == mock_device_constructor.return_value
        )
        assert (
            mock_devices_operations.create_or_update_identity.call_args[0][2]
            == '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager -- .get_device()")
class TestGetDevice(object):
    @pytest.mark.it("Gets device from service for provided device id")
    def test_get_device(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.get_device(fake_device_id)

        assert mock_devices_operations.get_identity.call_count == 1
        assert mock_devices_operations.get_identity.call_args == mocker.call(fake_device_id)


@pytest.mark.describe("IoTHubRegistryManager - .delete_device()")
class TestDeleteDevice(object):
    @pytest.mark.it("Deletes device for the provided device id")
    def test_delete_device(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.delete_device(fake_device_id)

        assert mock_devices_operations.delete_identity.call_count == 1
        assert mock_devices_operations.delete_identity.call_args == mocker.call(
            fake_device_id, '"*"'
        )

    @pytest.mark.it("Deletes device with an etag for the provided device id and etag")
    def test_delete_device_with_etag(
        self, mocker, mock_devices_operations, iothub_registry_manager
    ):
        iothub_registry_manager.delete_device(device_id=fake_device_id, etag=fake_etag)

        assert mock_devices_operations.delete_identity.call_count == 1
        assert mock_devices_operations.delete_identity.call_args == mocker.call(
            fake_device_id, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_module_with_sas()")
class TestCreateModuleWithSymmetricKey(object):

    testdata = [
        (fake_primary_key, None),
        (None, fake_secondary_key),
        (fake_primary_key, fake_secondary_key),
    ]

    @pytest.mark.it("Initializes module with device id, module id, managed_by and sas auth")
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_initializes_device_with_kwargs_for_sas(
        self, iothub_registry_manager, mock_module_constructor, primary_key, secondary_key
    ):
        iothub_registry_manager.create_module_with_sas(
            device_id=fake_device_id,
            module_id=fake_module_id,
            managed_by=fake_managed_by,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_module_constructor.call_count == 1

        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "sas"
        assert auth_mechanism.x509_thumbprint is None
        sym_key = auth_mechanism.symmetric_key
        assert sym_key.primary_key == primary_key
        assert sym_key.secondary_key == secondary_key

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key", "Both Keys"]
    )
    def test_calls_create_or_update_identity_for_sas(
        self,
        mock_module_constructor,
        mock_modules_operations,
        iothub_registry_manager,
        primary_key,
        secondary_key,
    ):
        iothub_registry_manager.create_module_with_sas(
            device_id=fake_device_id,
            module_id=fake_module_id,
            managed_by=fake_managed_by,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_module_with_x509()")
class TestCreateModuleWithX509(object):

    testdata = [
        (fake_primary_thumbprint, None),
        (None, fake_secondary_thumbprint),
        (fake_primary_thumbprint, fake_secondary_thumbprint),
    ]

    @pytest.mark.it("Initializes module with device id, module id, managed_by and X509 auth")
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_initializes_device_with_kwargs_for_x509(
        self,
        iothub_registry_manager,
        mock_module_constructor,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.create_module_with_x509(
            device_id=fake_device_id,
            module_id=fake_module_id,
            managed_by=fake_managed_by,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_module_constructor.call_count == 1
        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "selfSigned"
        assert auth_mechanism.symmetric_key is None
        x509_thumbprint = auth_mechanism.x509_thumbprint
        assert x509_thumbprint.primary_thumbprint == primary_thumbprint
        assert x509_thumbprint.secondary_thumbprint == secondary_thumbprint

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint", "Both Thumbprints"],
    )
    def test_calls_create_or_update_identity_for_x509(
        self,
        mock_module_constructor,
        mock_modules_operations,
        iothub_registry_manager,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.create_module_with_x509(
            device_id=fake_device_id,
            module_id=fake_module_id,
            managed_by=fake_managed_by,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .create_module_with_certificate_authority()")
class TestCreateModuleWithCA(object):
    @pytest.mark.it("Initializes module with device id, module id, managed_by and ca auth")
    def test_initializes_device_with_kwargs_for_certificate_authority(
        self, mock_module_constructor, iothub_registry_manager
    ):
        iothub_registry_manager.create_module_with_certificate_authority(
            device_id=fake_device_id, module_id=fake_module_id, managed_by=fake_managed_by
        )

        assert mock_module_constructor.call_count == 1
        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "certificateAuthority"
        assert auth_mechanism.x509_thumbprint is None
        assert auth_mechanism.symmetric_key is None

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    def test_calls_create_or_update_identity_for_certificate_authority(
        self, mock_module_constructor, mock_modules_operations, iothub_registry_manager
    ):
        iothub_registry_manager.create_module_with_certificate_authority(
            device_id=fake_device_id, module_id=fake_module_id, managed_by=fake_managed_by
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_module_with_sas()")
class TestUpdateModuleWithSymmetricKey(object):

    testdata = [(fake_primary_key, None), (None, fake_secondary_key)]

    @pytest.mark.it("Initializes module with device id, module id, managed_by and sas auth")
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key"]
    )
    def test_initializes_device_with_kwargs_for_sas(
        self, iothub_registry_manager, mock_module_constructor, primary_key, secondary_key
    ):
        iothub_registry_manager.update_module_with_sas(
            device_id=fake_device_id,
            module_id=fake_module_id,
            managed_by=fake_managed_by,
            etag=fake_etag,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_module_constructor.call_count == 1

        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "sas"
        assert auth_mechanism.x509_thumbprint is None
        sym_key = auth_mechanism.symmetric_key
        assert sym_key.primary_key == primary_key
        assert sym_key.secondary_key == secondary_key

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    @pytest.mark.parametrize(
        "primary_key, secondary_key", testdata, ids=["Primary Key", "Secondary Key"]
    )
    def test_calls_create_or_update_identity_for_sas(
        self,
        mock_module_constructor,
        mock_modules_operations,
        iothub_registry_manager,
        primary_key,
        secondary_key,
    ):
        iothub_registry_manager.update_module_with_sas(
            device_id=fake_device_id,
            module_id=fake_module_id,
            etag=fake_etag,
            managed_by=fake_managed_by,
            primary_key=primary_key,
            secondary_key=secondary_key,
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_module_with_x509()")
class TestUpdateModuleWithX509(object):

    testdata = [(fake_primary_thumbprint, None), (None, fake_secondary_thumbprint)]

    @pytest.mark.it("Initializes module with device id, module id, managed_by and X509 auth")
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint"],
    )
    def test_initializes_device_with_kwargs_for_x509(
        self,
        iothub_registry_manager,
        mock_module_constructor,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.update_module_with_x509(
            device_id=fake_device_id,
            module_id=fake_module_id,
            etag=fake_etag,
            managed_by=fake_managed_by,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_module_constructor.call_count == 1
        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "selfSigned"
        assert auth_mechanism.symmetric_key is None
        x509_thumbprint = auth_mechanism.x509_thumbprint
        assert x509_thumbprint.primary_thumbprint == primary_thumbprint
        assert x509_thumbprint.secondary_thumbprint == secondary_thumbprint

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    @pytest.mark.parametrize(
        "primary_thumbprint, secondary_thumbprint",
        testdata,
        ids=["Primary Thumbprint", "Secondary Thumbprint"],
    )
    def test_calls_create_or_update_identity_for_x509(
        self,
        mock_module_constructor,
        mock_modules_operations,
        iothub_registry_manager,
        primary_thumbprint,
        secondary_thumbprint,
    ):
        iothub_registry_manager.update_module_with_x509(
            device_id=fake_device_id,
            module_id=fake_module_id,
            etag=fake_etag,
            managed_by=fake_managed_by,
            primary_thumbprint=primary_thumbprint,
            secondary_thumbprint=secondary_thumbprint,
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_module_with_certificate_authority()")
class TestUpdateModuleWithCA(object):
    @pytest.mark.it("Initializes module with device id, module id, managed_by and ca auth")
    def test_initializes_device_with_kwargs_for_certificate_authority(
        self, mock_module_constructor, iothub_registry_manager
    ):
        iothub_registry_manager.update_module_with_certificate_authority(
            device_id=fake_device_id,
            module_id=fake_module_id,
            etag=fake_etag,
            managed_by=fake_managed_by,
        )

        assert mock_module_constructor.call_count == 1
        assert mock_module_constructor.call_args[1]["module_id"] == fake_module_id
        assert mock_module_constructor.call_args[1]["managed_by"] == fake_managed_by
        assert mock_module_constructor.call_args[1]["device_id"] == fake_device_id
        assert isinstance(
            mock_module_constructor.call_args[1]["authentication"], AuthenticationMechanism
        )
        auth_mechanism = mock_module_constructor.call_args[1]["authentication"]
        assert auth_mechanism.type == "certificateAuthority"
        assert auth_mechanism.x509_thumbprint is None
        assert auth_mechanism.symmetric_key is None

    @pytest.mark.it(
        "Calls method from service operations with device id, module id and previously constructed module"
    )
    def test_calls_create_or_update_identity_for_certificate_authority(
        self, mock_module_constructor, mock_modules_operations, iothub_registry_manager
    ):
        iothub_registry_manager.update_module_with_certificate_authority(
            device_id=fake_device_id,
            module_id=fake_module_id,
            etag=fake_etag,
            managed_by=fake_managed_by,
        )

        assert mock_modules_operations.create_or_update_identity.call_count == 1
        assert mock_modules_operations.create_or_update_identity.call_args[0][0] == fake_device_id
        assert mock_modules_operations.create_or_update_identity.call_args[0][1] == fake_module_id
        assert (
            mock_modules_operations.create_or_update_identity.call_args[0][2]
            == mock_module_constructor.return_value
        )


@pytest.mark.describe("IoTHubRegistryManager - .get_module()")
class TestGetModule(object):
    @pytest.mark.it("Gets module from service for provided device id and module id")
    def test_get_module(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.get_module(fake_device_id, fake_module_id)

        assert mock_modules_operations.get_identity.call_count == 1
        assert mock_modules_operations.get_identity.call_args == mocker.call(
            fake_device_id, fake_module_id
        )


@pytest.mark.describe("IoTHubRegistryManager - .get_modules()")
class TestGetModules(object):
    @pytest.mark.it("Gets all modules from service for provided device")
    def test_get_module(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.get_modules(fake_device_id)

        assert mock_modules_operations.get_modules_on_device.call_count == 1
        assert mock_modules_operations.get_modules_on_device.call_args == mocker.call(
            fake_device_id
        )


@pytest.mark.describe("IoTHubRegistryManager - .delete_module()")
class TestDeleteModule(object):
    @pytest.mark.it("Deletes module for the provided device id")
    def test_delete_module(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.delete_module(fake_device_id, fake_module_id)

        assert mock_modules_operations.delete_identity.call_count == 1
        assert mock_modules_operations.delete_identity.call_args == mocker.call(
            fake_device_id, fake_module_id, '"*"'
        )

    @pytest.mark.it("Deletes module with an etag for the provided device id and etag")
    def test_delete_module_with_etag(
        self, mocker, mock_modules_operations, iothub_registry_manager
    ):
        iothub_registry_manager.delete_module(
            device_id=fake_device_id, module_id=fake_module_id, etag=fake_etag
        )

        assert mock_modules_operations.delete_identity.call_count == 1
        assert mock_modules_operations.delete_identity.call_args == mocker.call(
            fake_device_id, fake_module_id, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .get_service_statistics()")
class TestGetServiceStats(object):
    @pytest.mark.it("Gets service statistics")
    def test_get_service_statistics(
        self, mocker, mock_statistics_operations, iothub_registry_manager
    ):
        iothub_registry_manager.get_service_statistics()

        assert mock_statistics_operations.get_service_statistics.call_count == 1
        assert mock_statistics_operations.get_service_statistics.call_args == mocker.call()


@pytest.mark.describe("IoTHubRegistryManager - .get_device_registry_statistics()")
class TestGetDeviceRegistryStats(object):
    @pytest.mark.it("Gets device registry statistics")
    def test_get_device_registry_statistics(
        self, mocker, mock_statistics_operations, iothub_registry_manager
    ):
        iothub_registry_manager.get_device_registry_statistics()

        assert mock_statistics_operations.get_device_statistics.call_count == 1
        assert mock_statistics_operations.get_device_statistics.call_args == mocker.call()


@pytest.mark.describe("IoTHubRegistryManager - .get_devices()")
class TestGetDevices(object):
    @pytest.mark.it("Gets devices")
    def test_get_devices(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.get_devices()

        assert mock_devices_operations.get_devices.call_count == 1
        assert mock_devices_operations.get_devices.call_args == mocker.call(None)


@pytest.mark.describe("IoTHubRegistryManager - .get_devices(max_number_of_devices)")
class TestGetDevicesWithMax(object):
    @pytest.mark.it("Gets devices with max_number_of_devices")
    def test_get_devices(self, mocker, mock_devices_operations, iothub_registry_manager):
        max_number_of_devices = 42
        iothub_registry_manager.get_devices(max_number_of_devices)

        assert mock_devices_operations.get_devices.call_count == 1
        assert mock_devices_operations.get_devices.call_args == mocker.call(max_number_of_devices)


@pytest.mark.describe("IoTHubRegistryManager - .bulk_create_or_update_identities()")
class TestBulkCreateUpdateDevices(object):
    @pytest.mark.it("Test bulk_create_or_update_identities")
    def test_bulk_create_or_update_identities(
        self, mocker, mock_bulk_registry_operations, iothub_registry_manager
    ):
        iothub_registry_manager.bulk_create_or_update_devices(fake_devices)
        assert mock_bulk_registry_operations.update_registry.call_count == 1
        assert mock_bulk_registry_operations.update_registry.call_args == mocker.call(fake_devices)


@pytest.mark.describe("IoTHubRegistryManager - .query_iot_hub()")
class TestQueryIoTHub(object):
    @pytest.mark.it("Test query IoTHub")
    def test_query_iot_hub(self, mocker, mock_query_operations, iothub_registry_manager):
        iothub_registry_manager.query_iot_hub(fake_query_specification)
        assert mock_query_operations.get_twins.call_count == 1
        assert mock_query_operations.get_twins.call_args == mocker.call(
            fake_query_specification, None, None, None, True
        )


@pytest.mark.describe("IoTHubRegistryManager - .query_iot_hub(continuation_token)")
class TestQueryIoTHubWithContinuationToken(object):
    @pytest.mark.it("Test query IoTHub with continuation token")
    def test_query_iot_hub(self, mocker, mock_query_operations, iothub_registry_manager):
        continuation_token = 42
        iothub_registry_manager.query_iot_hub(fake_query_specification, continuation_token)
        assert mock_query_operations.get_twins.call_count == 1
        assert mock_query_operations.get_twins.call_args == mocker.call(
            fake_query_specification, continuation_token, None, None, True
        )


@pytest.mark.describe("IoTHubRegistryManager - .query_iot_hub(continuation_token, max_item_count)")
class TestQueryIoTHubWithContinuationTokenAndMaxItermCount(object):
    @pytest.mark.it("Test query IoTHub with continuation token and max item count")
    def test_query_iot_hub(self, mocker, mock_query_operations, iothub_registry_manager):
        continuation_token = 42
        max_item_count = 84
        iothub_registry_manager.query_iot_hub(
            fake_query_specification, continuation_token, max_item_count
        )
        assert mock_query_operations.get_twins.call_count == 1
        assert mock_query_operations.get_twins.call_args == mocker.call(
            fake_query_specification, continuation_token, max_item_count, None, True
        )


@pytest.mark.describe("IoTHubRegistryManager - .get_twin()")
class TestGetTwin(object):
    @pytest.mark.it("Test get twin")
    def test_get_twin(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.get_twin(fake_device_id)
        assert mock_devices_operations.get_twin.call_count == 1
        assert mock_devices_operations.get_twin.call_args == mocker.call(fake_device_id)


@pytest.mark.describe("IoTHubRegistryManager - .replace_twin()")
class TestReplaceTwin(object):
    @pytest.mark.it("Test replace twin")
    def test_replace_twin(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.replace_twin(fake_device_id, fake_device_twin, fake_etag)
        assert mock_devices_operations.replace_twin.call_count == 1
        assert mock_devices_operations.replace_twin.call_args == mocker.call(
            fake_device_id, fake_device_twin, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_twin()")
class TestUpdateTwin(object):
    @pytest.mark.it("Test update twin")
    def test_update_twin(self, mocker, mock_devices_operations, iothub_registry_manager):
        iothub_registry_manager.update_twin(fake_device_id, fake_device_twin, fake_etag)
        assert mock_devices_operations.update_twin.call_count == 1
        assert mock_devices_operations.update_twin.call_args == mocker.call(
            fake_device_id, fake_device_twin, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .get_module_twin()")
class TestGetModuleTwin(object):
    @pytest.mark.it("Test get module twin")
    def test_get_module_twin(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.get_module_twin(fake_device_id, fake_module_id)
        assert mock_modules_operations.get_twin.call_count == 1
        assert mock_modules_operations.get_twin.call_args == mocker.call(
            fake_device_id, fake_module_id
        )


@pytest.mark.describe("IoTHubRegistryManager - .replace_module_twin()")
class TestReplaceModuleTwin(object):
    @pytest.mark.it("Test replace module twin")
    def test_replace_module_twin(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.replace_module_twin(
            fake_device_id, fake_module_id, fake_module_twin, fake_etag
        )
        assert mock_modules_operations.replace_twin.call_count == 1
        assert mock_modules_operations.replace_twin.call_args == mocker.call(
            fake_device_id, fake_module_id, fake_module_twin, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .update_module_twin()")
class TestUpdateModuleTwin(object):
    @pytest.mark.it("Test update module twin")
    def test_update_module_twin(self, mocker, mock_modules_operations, iothub_registry_manager):
        iothub_registry_manager.update_module_twin(
            fake_device_id, fake_module_id, fake_module_twin, fake_etag
        )
        assert mock_modules_operations.update_twin.call_count == 1
        assert mock_modules_operations.update_twin.call_args == mocker.call(
            fake_device_id, fake_module_id, fake_module_twin, '"' + fake_etag + '"'
        )


@pytest.mark.describe("IoTHubRegistryManager - .invoke_device_method() with payload")
class TestInvokeDeviceMethodWithPayload(object):
    @pytest.mark.it("Test invoke device method with payload")
    def test_invoke_device_method_with_payload(
        self, mocker, mock_devices_operations, iothub_registry_manager
    ):
        iothub_registry_manager.invoke_device_method(
            fake_device_id, fake_direct_method_request_with_payload
        )
        assert mock_devices_operations.invoke_method.call_count == 1
        assert mock_devices_operations.invoke_method.call_args == mocker.call(
            fake_device_id, fake_direct_method_request_with_payload
        )


@pytest.mark.describe("IoTHubRegistryManager - .invoke_device_method() with payload none")
class TestInvokeDeviceMethodWithPayloadNone(object):
    @pytest.mark.it("Test invoke device method with payload None")
    def test_invoke_device_method_payload_none(
        self, mocker, mock_devices_operations, iothub_registry_manager
    ):
        iothub_registry_manager.invoke_device_method(
            fake_device_id, fake_direct_method_request_with_payload_none
        )
        assert mock_devices_operations.invoke_method.call_count == 1
        assert mock_devices_operations.invoke_method.call_args == mocker.call(
            fake_device_id, fake_direct_method_request_with_payload_none
        )


@pytest.mark.describe("IoTHubRegistryManager - .invoke_device_module_method() with payload")
class TestInvokeDeviceModuleMethodWithPayload(object):
    @pytest.mark.it("Test invoke device module method with payload")
    def test_invoke_device_module_method_with_payload(
        self, mocker, mock_modules_operations, iothub_registry_manager
    ):
        iothub_registry_manager.invoke_device_module_method(
            fake_device_id, fake_module_id, fake_direct_method_request_with_payload
        )
        assert mock_modules_operations.invoke_method.call_count == 1
        assert mock_modules_operations.invoke_method.call_args == mocker.call(
            fake_device_id, fake_module_id, fake_direct_method_request_with_payload
        )


@pytest.mark.describe("IoTHubRegistryManager - .invoke_device_module_method() with payload none")
class TestInvokeDeviceModuleMethodWithPayloadNone(object):
    @pytest.mark.it("Test invoke device module method with payload None")
    def test_invoke_device_module_method_payload_none(
        self, mocker, mock_modules_operations, iothub_registry_manager
    ):
        iothub_registry_manager.invoke_device_module_method(
            fake_device_id, fake_module_id, fake_direct_method_request_with_payload_none
        )
        assert mock_modules_operations.invoke_method.call_count == 1
        assert mock_modules_operations.invoke_method.call_args == mocker.call(
            fake_device_id, fake_module_id, fake_direct_method_request_with_payload_none
        )


@pytest.mark.describe("IoTHubRegistryManager - .send_c2d_message()")
class TestSendC2dMessage(object):
    @pytest.mark.it("Test send c2d message")
    def test_send_c2d_message(
        self, mocker, mock_uamqp_send_message_to_device, iothub_registry_manager
    ):

        iothub_registry_manager.send_c2d_message(fake_device_id, fake_message_to_send)

        assert mock_uamqp_send_message_to_device.call_count == 1
        assert mock_uamqp_send_message_to_device.call_args == mocker.call(
            fake_device_id, fake_message_to_send, {}
        )


@pytest.mark.describe("IoTHubRegistryManager - .send_c2d_message() with properties")
class TestSendC2dMessageWithProperties(object):
    @pytest.mark.it("Test send c2d message with properties")
    def test_send_c2d_message_with_properties(
        self, mocker, mock_uamqp_send_message_to_device, iothub_registry_manager
    ):

        iothub_registry_manager.send_c2d_message(
            fake_device_id, fake_message_to_send, fake_properties
        )

        assert mock_uamqp_send_message_to_device.call_count == 1
        assert mock_uamqp_send_message_to_device.call_args == mocker.call(
            fake_device_id, fake_message_to_send, fake_properties
        )



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.hub.iothub_job_manager import IoTHubJobManager
from azure.iot.hub.auth import ConnectionStringAuthentication
from azure.iot.hub.protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs

"""---Constants---"""

fake_hostname = "beauxbatons.academy-net"
fake_device_id = "MyPensieve"
fake_shared_access_key_name = "alohomora"
fake_shared_access_key = "Zm9vYmFy"
fake_job_properties = "fake_job_properties"
fake_job_id = "fake_job_id"
fake_job_request = "fake_job_request"
fake_job_type = "fake_job_type"
fake_job_status = "fake_job_status"


"""----Shared fixtures----"""


@pytest.fixture(scope="function", autouse=True)
def mock_job_client_operations(mocker):
    mock_job_client_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.JobsOperations"
    )
    return mock_job_client_operations_init.return_value


@pytest.fixture(scope="function")
def iothub_job_manager():
    connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
        hostname=fake_hostname,
        device_id=fake_device_id,
        skn=fake_shared_access_key_name,
        sk=fake_shared_access_key,
    )
    iothub_job_manager = IoTHubJobManager.from_connection_string(connection_string)
    return iothub_job_manager


@pytest.mark.describe("IoTHubJobManager - .from_connection_string()")
class TestFromConnectionString:
    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
                    hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
                ),
                id="connection string without SharedAccessKeyName",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of ConnectionStringAuthentication and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_connection_string_auth(self, connection_string):
        client = IoTHubJobManager.from_connection_string(connection_string=connection_string)

        assert repr(client.auth) == connection_string
        assert client.protocol.config.base_url == "https://" + client.auth["HostName"]
        assert client.protocol.config.credentials == client.auth

    @pytest.mark.it("Sets the auth and protocol attributes")
    def test_instantiates_auth_and_protocol_attributes(self, iothub_job_manager):
        assert isinstance(iothub_job_manager.auth, ConnectionStringAuthentication)
        assert isinstance(iothub_job_manager.protocol, IotHubGatewayServiceAPIs)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with an empty connection string"
    )
    def test_instantiates_with_empty_connection_string(self):
        with pytest.raises(ValueError):
            IoTHubJobManager.from_connection_string("")

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without HostName"
    )
    def test_instantiates_with_connection_string_no_host_name(self):
        connection_string = (
            "DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                device_id=fake_device_id, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        with pytest.raises(ValueError):
            IoTHubJobManager.from_connection_string(connection_string)

    @pytest.mark.it("Instantiates with an connection string without DeviceId")
    def test_instantiates_with_connection_string_no_device_id(self):
        connection_string = (
            "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                hostname=fake_hostname, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        obj = IoTHubJobManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubJobManager)

    @pytest.mark.it("Instantiates with an connection string without SharedAccessKeyName")
    def test_instantiates_with_connection_string_no_shared_access_key_name(self):
        connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
            hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
        )
        obj = IoTHubJobManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubJobManager)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without SharedAccessKey"
    )
    def test_instantiates_with_connection_string_no_shared_access_key(self):
        connection_string = (
            "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn}".format(
                hostname=fake_hostname, device_id=fake_device_id, skn=fake_shared_access_key_name
            )
        )
        with pytest.raises(ValueError):
            IoTHubJobManager.from_connection_string(connection_string)


@pytest.mark.describe("IoTHubJobManager - .from_token_credential()")
class TestFromTokenCredential:
    @pytest.mark.it(
        "Creates an instance of AzureIdentityCredentialAdapter and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_token_credential_auth(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()

        client = IoTHubJobManager.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential
        )

        assert client.auth._policy._credential == mock_azure_identity_TokenCredential
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert client.protocol.config.credentials == client.auth


@pytest.mark.describe("IoTHubJobManager - .create_import_export_job()")
class TestCreateImportExportJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to create an export/import job")
    def test_create_export_import_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.create_import_export_job(fake_job_properties)
        assert mock_job_client_operations.create_import_export_job.call_count == 1
        assert mock_job_client_operations.create_import_export_job.call_args == mocker.call(
            fake_job_properties
        )
        assert ret_val == mock_job_client_operations.create_import_export_job()


@pytest.mark.describe("IoTHubJobManager - .get_import_export_jobs()")
class TestGetImportExportJobs(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to get an export/import jobs")
    def test_get_export_import_jobs(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.get_import_export_jobs()
        assert mock_job_client_operations.get_import_export_jobs.call_count == 1
        assert mock_job_client_operations.get_import_export_jobs.call_args == mocker.call()
        assert ret_val == mock_job_client_operations.get_import_export_jobs()


@pytest.mark.describe("IoTHubJobManager - .get_import_export_job()")
class TestGetImportExportJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to get an export/import job")
    def test_get_export_import_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.get_import_export_job(fake_job_id)
        assert mock_job_client_operations.get_import_export_job.call_count == 1
        assert mock_job_client_operations.get_import_export_job.call_args == mocker.call(
            fake_job_id
        )
        assert ret_val == mock_job_client_operations.get_import_export_job()


@pytest.mark.describe("IoTHubJobManager - .cancel_import_export_job()")
class TestCancelImportExportJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to cancel an export/import job")
    def test_cancel_import_export_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.cancel_import_export_job(fake_job_id)
        assert mock_job_client_operations.cancel_import_export_job.call_count == 1
        assert mock_job_client_operations.cancel_import_export_job.call_args == mocker.call(
            fake_job_id
        )
        assert ret_val == mock_job_client_operations.cancel_import_export_job()


@pytest.mark.describe("IoTHubJobManager - .create_job()")
class TestCreateScheduledJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to create a job")
    def test_create_scheduled_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.create_scheduled_job(fake_job_id, fake_job_request)
        assert mock_job_client_operations.create_scheduled_job.call_count == 1
        assert mock_job_client_operations.create_scheduled_job.call_args == mocker.call(
            fake_job_id, fake_job_request
        )
        assert ret_val == mock_job_client_operations.create_scheduled_job()


@pytest.mark.describe("IoTHubJobManager - .get_job()")
class TestGetScheduledJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to get a job")
    def test_get_scheduled_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.get_scheduled_job(fake_job_id)
        assert mock_job_client_operations.get_scheduled_job.call_count == 1
        assert mock_job_client_operations.get_scheduled_job.call_args == mocker.call(fake_job_id)
        assert ret_val == mock_job_client_operations.get_scheduled_job()


@pytest.mark.describe("IoTHubJobManager - .cancel_job()")
class TestCancelScheduledJob(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to cancel a job")
    def test_cancel_scheduled_job(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.cancel_scheduled_job(fake_job_id)
        assert mock_job_client_operations.cancel_scheduled_job.call_count == 1
        assert mock_job_client_operations.cancel_scheduled_job.call_args == mocker.call(fake_job_id)
        assert ret_val == mock_job_client_operations.cancel_scheduled_job()


@pytest.mark.describe("IoTHubJobManager - .query_jobs()")
class TestQueryScheduledJobs(object):
    @pytest.mark.it("Uses protocol layer Job Client runtime to query a job")
    def test_query_scheduled_jobs(self, mocker, mock_job_client_operations, iothub_job_manager):
        ret_val = iothub_job_manager.query_scheduled_jobs(fake_job_type, fake_job_status)
        assert mock_job_client_operations.query_scheduled_jobs.call_count == 1
        assert mock_job_client_operations.query_scheduled_jobs.call_args == mocker.call(
            fake_job_type, fake_job_status
        )
        assert ret_val == mock_job_client_operations.query_scheduled_jobs()



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import base64
import hashlib
import pytest
import uamqp
import hmac
from azure.core.credentials import AccessToken
from azure.iot.hub.iothub_amqp_client import (
    IoTHubAmqpClientSharedAccessKeyAuth,
    IoTHubAmqpClientTokenAuth,
)

"""---Constants---"""
fake_device_id = "device_id"
fake_message = "fake_message"
fake_app_prop = {"fake_prop1": "fake_value1"}
fake_sys_prop = {
    "contentType": "value1",
    "contentEncoding": "value2",
    "correlationId": "value3",
    "expiryTimeUtc": 1584727659,
    "messageId": 42,
}
fake_hostname = "hostname.mytest.net"
fake_shared_access_key_name = "test_key_name"
fake_shared_access_key = "Zm9vYmFy"
fake_token_scope = "https://token.net/.default"
fake_token = "fasdjkhg"
fake_token_expiry = 1584727659


@pytest.fixture(autouse=True)
def mock_uamqp_SendClient(mocker):
    mock_uamqp_SendClient = mocker.patch.object(uamqp, "SendClient")
    return mock_uamqp_SendClient


class SharedIotHubAmqpClientSendMessageToDeviceTests(object):
    @pytest.mark.it("Sends messages with application properties using the uamqp SendClient")
    def test_send_message_to_device_app_prop(self, client, mocker, mock_uamqp_SendClient):
        client.send_message_to_device(fake_device_id, fake_message, fake_app_prop)
        amqp_client_obj = mock_uamqp_SendClient.return_value

        # Message was queued
        assert amqp_client_obj.queue_message.call_count == 1
        msg_obj = amqp_client_obj.queue_message.call_args[0][0]
        # Message was configured with properties and destination
        assert str(msg_obj) == fake_message
        assert msg_obj.properties.to == bytes(
            str("/devices/" + fake_device_id + "/messages/devicebound").encode("utf-8")
        )
        assert msg_obj.application_properties == fake_app_prop
        # Message was sent
        assert amqp_client_obj.send_all_messages.call_count == 1
        assert amqp_client_obj.send_all_messages.call_args == mocker.call(close_on_done=False)

    @pytest.mark.it("Sends messages with system properties using the uamqp SendClient")
    def test_send_message_to_device_sys_prop(self, client, mocker, mock_uamqp_SendClient):
        client.send_message_to_device(fake_device_id, fake_message, fake_sys_prop)
        amqp_client_obj = mock_uamqp_SendClient.return_value

        # Message was queued
        assert amqp_client_obj.queue_message.call_count == 1
        msg_obj = amqp_client_obj.queue_message.call_args[0][0]
        # Message was configured with properties and destination
        assert str(msg_obj) == fake_message
        assert msg_obj.properties.to == bytes(
            str("/devices/" + fake_device_id + "/messages/devicebound").encode("utf-8")
        )
        assert msg_obj.application_properties == {}
        assert msg_obj.properties.content_type == bytes(
            str(fake_sys_prop["contentType"]).encode("utf-8")
        )
        assert msg_obj.properties.content_encoding == bytes(
            str(fake_sys_prop["contentEncoding"]).encode("utf-8")
        )
        assert msg_obj.properties.correlation_id == bytes(
            str(fake_sys_prop["correlationId"]).encode("utf-8")
        )
        assert msg_obj.properties.absolute_expiry_time == fake_sys_prop["expiryTimeUtc"]
        assert msg_obj.properties.message_id == fake_sys_prop["messageId"]
        # Message was sent
        assert amqp_client_obj.send_all_messages.call_count == 1
        assert amqp_client_obj.send_all_messages.call_args == mocker.call(close_on_done=False)

    @pytest.mark.it("Raises an Exception if send_all_messages fails")
    def test_raise_exception_on_send_fail(self, client, mocker, mock_uamqp_SendClient):
        amqp_client_obj = mock_uamqp_SendClient.return_value
        mocker.patch.object(
            amqp_client_obj, "send_all_messages", {uamqp.constants.MessageState.SendFailed}
        )
        with pytest.raises(Exception):
            client.send_message_to_device(fake_device_id, fake_message, fake_app_prop)


class SharedIotHubAmqpClientDisconnectSyncTests(object):
    @pytest.mark.it(
        "Calls close() on the uamqp SendClient and removes it when disconnect_sync() is called"
    )
    def test_disconnect_sync(self, client, mock_uamqp_SendClient):
        assert client.amqp_client is not None
        amqp_client_obj = mock_uamqp_SendClient.return_value

        client.disconnect_sync()

        assert amqp_client_obj.close.call_count == 1
        assert client.amqp_client is None


#############################################
# IoTHubAmqpClientSharedAccessKeyAuth Tests #
#############################################


class IoTHubAmqpClientSharedAccessKeyAuthTestConfig(object):
    @pytest.fixture
    def client(self):
        return IoTHubAmqpClientSharedAccessKeyAuth(
            hostname=fake_hostname,
            shared_access_key_name=fake_shared_access_key_name,
            shared_access_key=fake_shared_access_key,
        )


@pytest.mark.describe("IoTHubAmqpClientSharedAccessKeyAuth - Instantiation")
class TestIoTHubAmqpClientSharedAccessKeyAuthInstantiation(
    IoTHubAmqpClientSharedAccessKeyAuthTestConfig
):
    @pytest.mark.it(
        "Creates a JWTTokenAuth instance with the correct parameters and uses it to create an AMQP SendClient"
    )
    def test_create_JWTTokenAuth_with_sas_token(self, mocker, mock_uamqp_SendClient):
        amqp_token_init_mock = mocker.patch.object(uamqp.authentication, "JWTTokenAuth")
        amqp_token_mock = amqp_token_init_mock.return_value

        IoTHubAmqpClientSharedAccessKeyAuth(
            fake_hostname, fake_shared_access_key_name, fake_shared_access_key
        )

        # JWTTokenAuth creation
        assert amqp_token_init_mock.call_count == 1
        assert amqp_token_init_mock.call_args[1]["uri"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["audience"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["token_type"] == b"servicebus.windows.net:sastoken"
        assert amqp_token_init_mock.call_args[1]["transport_type"] == uamqp.TransportType.Amqp
        assert amqp_token_mock.update_token.call_count == 1

        # AMQP SendClient is created
        assert mock_uamqp_SendClient.call_count == 1
        expected_target = "amqps://" + fake_hostname + "/messages/devicebound"
        assert mock_uamqp_SendClient.call_args == mocker.call(
            target=expected_target, auth=amqp_token_mock, keep_alive_interval=120, transport_type=uamqp.TransportType.Amqp
        )

    @pytest.mark.it(
        "Creates a JWTTokenAuth instance with the correct parameters and uses it to create an AMQP over Websocket SendClient"
    )
    def test_create_JWTTokenAuth_with_sas_token_amqp_over_websocket(self, mocker, mock_uamqp_SendClient):
        amqp_token_init_mock = mocker.patch.object(uamqp.authentication, "JWTTokenAuth")
        amqp_token_mock = amqp_token_init_mock.return_value

        IoTHubAmqpClientSharedAccessKeyAuth(
            fake_hostname, fake_shared_access_key_name, fake_shared_access_key, transport_type=uamqp.TransportType.AmqpOverWebsocket
        )

        # JWTTokenAuth creation
        assert amqp_token_init_mock.call_count == 1
        assert amqp_token_init_mock.call_args[1]["uri"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["audience"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["token_type"] == b"servicebus.windows.net:sastoken"
        assert amqp_token_init_mock.call_args[1]["transport_type"] == uamqp.TransportType.AmqpOverWebsocket
        assert amqp_token_mock.update_token.call_count == 1

        # AMQP over Websocket SendClient is created
        assert mock_uamqp_SendClient.call_count == 1
        expected_target = "amqps://" + fake_hostname + "/messages/devicebound"
        assert mock_uamqp_SendClient.call_args == mocker.call(
            target=expected_target, auth=amqp_token_mock, keep_alive_interval=120, transport_type=uamqp.TransportType.AmqpOverWebsocket
        )

    @pytest.mark.it("Creates an HMAC to generate a shared access signature")
    def test_creates_hmac(self, mocker):
        hmac_mock = mocker.patch.object(hmac, "HMAC")
        hmac_digest_mock = hmac_mock.return_value.digest
        hmac_digest_mock.return_value = b"\xd2\x06\xf7\x12\xf1\xe9\x95$\x90\xfd\x12\x9a\xb1\xbe\xb4\xf8\xf3\xc4\x1ap\x8a\xab'\x8a.D\xfb\x84\x96\xca\xf3z"

        IoTHubAmqpClientSharedAccessKeyAuth(
            fake_hostname, fake_shared_access_key_name, fake_shared_access_key
        )

        assert hmac_mock.call_count == 1
        assert hmac_mock.call_args == mocker.call(
            base64.b64decode(fake_shared_access_key + "="), mocker.ANY, hashlib.sha256
        )

        assert hmac_digest_mock.call_count == 1


@pytest.mark.describe("IoTHubAmqpClientSharedAccessKeyAuth - .send_message_to_device()")
class TestIoTHubAmqpClientSharedAccessKeyAuthSendMessageToDevice(
    IoTHubAmqpClientSharedAccessKeyAuthTestConfig, SharedIotHubAmqpClientSendMessageToDeviceTests
):
    pass


@pytest.mark.describe("IoTHubAmqpClientSharedAccessKeyAuth - .disconnect_sync()")
class TestIoTHubAmqpClientSharedAccessKeyAuthDisconnectSync(
    IoTHubAmqpClientSharedAccessKeyAuthTestConfig, SharedIotHubAmqpClientDisconnectSyncTests
):
    pass


###################################
# IoTHubAmqpClientTokenAuth Tests #
###################################


class IoTHubAmqpClientTokenAuthTestConfig(object):
    @pytest.fixture
    def mock_azure_identity_TokenCredential(self, mocker):
        mock = mocker.MagicMock()
        mock.get_token.return_value = AccessToken(fake_token, fake_token_expiry)
        return mock

    @pytest.fixture
    def client(self, mock_azure_identity_TokenCredential):
        return IoTHubAmqpClientTokenAuth(
            hostname=fake_hostname,
            token_credential=mock_azure_identity_TokenCredential,
            token_scope=fake_token_scope,
        )


@pytest.mark.describe("IoTHubAmqpClientTokenAuth - Instantiation")
class TestIotHubAmqpClientTokenAuthInstantiation(IoTHubAmqpClientTokenAuthTestConfig):
    @pytest.mark.it(
        "Creates a JWTTokenAuth instance with the correct parameters and uses it to create an AMQP SendClient when a token scope is specified"
    )
    def test_create_JWTTokenAuth_with_bearer_token_custom_scope(
        self, mocker, mock_azure_identity_TokenCredential, mock_uamqp_SendClient
    ):
        amqp_token_init_mock = mocker.patch.object(uamqp.authentication, "JWTTokenAuth")
        amqp_token_mock = amqp_token_init_mock.return_value

        IoTHubAmqpClientTokenAuth(
            fake_hostname, mock_azure_identity_TokenCredential, token_scope=fake_token_scope
        )

        # JWTTokenAuth Creation
        assert amqp_token_init_mock.call_count == 1
        assert amqp_token_init_mock.call_args[1]["uri"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["audience"] == fake_token_scope
        assert amqp_token_init_mock.call_args[1]["token_type"] == b"bearer"
        assert amqp_token_init_mock.call_args[1]["transport_type"] == uamqp.TransportType.Amqp
        assert amqp_token_mock.update_token.call_count == 1

        # AMQP SendClient is created
        assert mock_uamqp_SendClient.call_count == 1
        expected_target = "amqps://" + fake_hostname + "/messages/devicebound"
        assert mock_uamqp_SendClient.call_args == mocker.call(
            target=expected_target, auth=amqp_token_mock, keep_alive_interval=120, transport_type=uamqp.TransportType.Amqp
        )

    @pytest.mark.it(
        "Creates a JWTTokenAuth instance with the correct parameters and uses it to create an AMQP SendClient if no token scope is provided (uses the default scope of 'https://iothubs.azure.net/.default')"
    )
    def test_create_JWTTokenAuth_with_bearer_token_default_scope(
        self, mocker, mock_azure_identity_TokenCredential, mock_uamqp_SendClient
    ):
        amqp_token_init_mock = mocker.patch.object(uamqp.authentication, "JWTTokenAuth")
        amqp_token_mock = amqp_token_init_mock.return_value

        IoTHubAmqpClientTokenAuth(fake_hostname, mock_azure_identity_TokenCredential)

        # JWTTokenAuth Creation
        assert amqp_token_init_mock.call_count == 1
        assert amqp_token_init_mock.call_args[1]["uri"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["audience"] == "https://iothubs.azure.net/.default"
        assert amqp_token_init_mock.call_args[1]["token_type"] == b"bearer"
        assert amqp_token_init_mock.call_args[1]["transport_type"] == uamqp.TransportType.Amqp
        assert amqp_token_mock.update_token.call_count == 1

        # AMQP SendClient is created
        assert mock_uamqp_SendClient.call_count == 1
        expected_target = "amqps://" + fake_hostname + "/messages/devicebound"
        assert mock_uamqp_SendClient.call_args == mocker.call(
            target=expected_target, auth=amqp_token_mock, keep_alive_interval=120, transport_type=uamqp.TransportType.Amqp
        )

    @pytest.mark.it(
        "Creates a JWTTokenAuth instance with the correct parameters and uses it to create an AMQP over WebSocket SendClient if no token scope is provided (uses the default scope of 'https://iothubs.azure.net/.default')"
    )
    def test_create_JWTTokenAuth_with_bearer_token_default_scope_AmqpOverWebsocket(
        self, mocker, mock_azure_identity_TokenCredential, mock_uamqp_SendClient
    ):
        amqp_token_init_mock = mocker.patch.object(uamqp.authentication, "JWTTokenAuth")
        amqp_token_mock = amqp_token_init_mock.return_value

        IoTHubAmqpClientTokenAuth(fake_hostname, mock_azure_identity_TokenCredential, transport_type=uamqp.TransportType.AmqpOverWebsocket)

        # JWTTokenAuth Creation
        assert amqp_token_init_mock.call_count == 1
        assert amqp_token_init_mock.call_args[1]["uri"] == "https://" + fake_hostname
        assert amqp_token_init_mock.call_args[1]["audience"] == "https://iothubs.azure.net/.default"
        assert amqp_token_init_mock.call_args[1]["token_type"] == b"bearer"
        assert amqp_token_init_mock.call_args[1]["transport_type"] == uamqp.TransportType.AmqpOverWebsocket
        assert amqp_token_mock.update_token.call_count == 1

        # AMQP over Websocket SendClient is created
        assert mock_uamqp_SendClient.call_count == 1
        expected_target = "amqps://" + fake_hostname + "/messages/devicebound"
        assert mock_uamqp_SendClient.call_args == mocker.call(
            target=expected_target, auth=amqp_token_mock, keep_alive_interval=120, transport_type=uamqp.TransportType.AmqpOverWebsocket
        )

    @pytest.mark.it(
        "Retrieves the token from the azure-identity TokenCredential using the specified token scope"
    )
    def test_retrieve_token_from_azure_identity(self, mock_azure_identity_TokenCredential):
        IoTHubAmqpClientTokenAuth(
            fake_hostname, mock_azure_identity_TokenCredential, fake_token_scope
        )
        mock_azure_identity_TokenCredential.get_token.assert_called_once_with(fake_token_scope)


@pytest.mark.describe("IoTHubAmqpClientTokenAuth - .send_message_to_device()")
class TestIoTHubAmqpClientTokenAuthSendMessageToDevice(
    IoTHubAmqpClientTokenAuthTestConfig, SharedIotHubAmqpClientSendMessageToDeviceTests
):
    pass


@pytest.mark.describe("IoTHubAmqpClientTokenAuth - .disconnect_sync()")
class TestIoTHubAmqpClientTokenAuthDisconnectSync(
    IoTHubAmqpClientTokenAuthTestConfig, SharedIotHubAmqpClientDisconnectSyncTests
):
    pass



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.hub.digital_twin_client import DigitalTwinClient
from azure.iot.hub.auth import ConnectionStringAuthentication
from azure.iot.hub.protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs

"""---Constants---"""

fake_hostname = "beauxbatons.academy-net"
fake_device_id = "MyPensieve"
fake_shared_access_key_name = "alohomora"
fake_shared_access_key = "Zm9vYmFy"
fake_digital_twin_id = "fake_digital_twin_id"
fake_digital_twin_patch = "fake_digital_twin_patch"
fake_etag = "fake_etag"
fake_component_path = "fake_component_path"
fake_component_name = "fake_component_name"
fake_payload = "fake_payload"
fake_model_id = "fake_model_id"
fake_conn_timeout = 42
fake_resp_timeout = 72


"""----Shared fixtures----"""


@pytest.fixture(scope="function", autouse=True)
def mock_digital_twin_operations(mocker):
    mock_digital_twin_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.DigitalTwinOperations"
    )
    return mock_digital_twin_operations_init.return_value


@pytest.fixture(scope="function")
def digital_twin_client():
    connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
        hostname=fake_hostname,
        device_id=fake_device_id,
        skn=fake_shared_access_key_name,
        sk=fake_shared_access_key,
    )
    digital_twin_client = DigitalTwinClient.from_connection_string(connection_string)
    return digital_twin_client


@pytest.mark.describe("DigitalTwinClient - .from_connection_string()")
class TestFromConnectionString:
    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
                    hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
                ),
                id="connection string without SharedAccessKeyName",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of ConnectionStringAuthentication and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_connection_string_auth(self, connection_string):
        client = DigitalTwinClient.from_connection_string(connection_string=connection_string)

        assert repr(client.auth) == connection_string
        assert client.protocol.config.base_url == "https://" + client.auth["HostName"]
        assert client.protocol.config.credentials == client.auth

    @pytest.mark.it("Sets the auth and protocol attributes")
    def test_instantiates_auth_and_protocol_attributes(self, digital_twin_client):
        assert isinstance(digital_twin_client.auth, ConnectionStringAuthentication)
        assert isinstance(digital_twin_client.protocol, IotHubGatewayServiceAPIs)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with an empty connection string"
    )
    def test_instantiates_with_empty_connection_string(self):
        with pytest.raises(ValueError):
            DigitalTwinClient.from_connection_string("")

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without HostName"
    )
    def test_instantiates_with_connection_string_no_host_name(self):
        connection_string = (
            "DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                device_id=fake_device_id, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        with pytest.raises(ValueError):
            DigitalTwinClient.from_connection_string(connection_string)

    @pytest.mark.it("Instantiates with an connection string without DeviceId")
    def test_instantiates_with_connection_string_no_device_id(self):
        connection_string = (
            "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                hostname=fake_hostname, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        obj = DigitalTwinClient.from_connection_string(connection_string)
        assert isinstance(obj, DigitalTwinClient)

    @pytest.mark.it("Instantiates with an connection string without SharedAccessKeyName")
    def test_instantiates_with_connection_string_no_shared_access_key_name(self):
        connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
            hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
        )
        obj = DigitalTwinClient.from_connection_string(connection_string)
        assert isinstance(obj, DigitalTwinClient)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without SharedAccessKey"
    )
    def test_instantiates_with_connection_string_no_shared_access_key(self):
        connection_string = (
            "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn}".format(
                hostname=fake_hostname, device_id=fake_device_id, skn=fake_shared_access_key_name
            )
        )
        with pytest.raises(ValueError):
            DigitalTwinClient.from_connection_string(connection_string)


@pytest.mark.describe("DigitalTwinClient - .from_token_credential()")
class TestFromTokenCredential:
    @pytest.mark.it(
        "Creates an instance of AzureIdentityCredentialAdapter and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_token_credential_auth(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()

        client = DigitalTwinClient.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential
        )

        assert client.auth._policy._credential == mock_azure_identity_TokenCredential
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert client.protocol.config.credentials == client.auth


@pytest.mark.describe("DigitalTwinClient - .get_digital_twin()")
class TestGetDigitalTwin(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to get a digital twin")
    def test_get_digital_twin(self, mocker, mock_digital_twin_operations, digital_twin_client):
        ret_val = digital_twin_client.get_digital_twin(fake_digital_twin_id)
        assert mock_digital_twin_operations.get_digital_twin.call_count == 1
        assert mock_digital_twin_operations.get_digital_twin.call_args == mocker.call(
            fake_digital_twin_id
        )
        assert ret_val == mock_digital_twin_operations.get_digital_twin()


@pytest.mark.describe("DigitalTwinClient - .update_digital_twin()")
class TestUpdateDigitalTwin(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to update a digital twin")
    def test_update_digital_twin(self, mocker, mock_digital_twin_operations, digital_twin_client):
        ret_val = digital_twin_client.update_digital_twin(
            fake_digital_twin_id, fake_digital_twin_patch, fake_etag
        )
        assert mock_digital_twin_operations.update_digital_twin.call_count == 1
        assert mock_digital_twin_operations.update_digital_twin.call_args == mocker.call(
            fake_digital_twin_id, fake_digital_twin_patch, fake_etag
        )
        assert ret_val == mock_digital_twin_operations.update_digital_twin()


@pytest.mark.describe("DigitalTwinClient - .update_digital_twin()")
class TestUpdateDigitalTwinNoEtag(object):
    @pytest.mark.it(
        "Uses protocol layer DigitalTwin Client runtime to update a digital twin without etag"
    )
    def test_update_digital_twin(self, mocker, mock_digital_twin_operations, digital_twin_client):
        ret_val = digital_twin_client.update_digital_twin(
            fake_digital_twin_id, fake_digital_twin_patch
        )
        assert mock_digital_twin_operations.update_digital_twin.call_count == 1
        assert mock_digital_twin_operations.update_digital_twin.call_args == mocker.call(
            fake_digital_twin_id, fake_digital_twin_patch, None
        )
        assert ret_val == mock_digital_twin_operations.update_digital_twin()


@pytest.mark.describe("DigitalTwinClient - .invoke_component_command()")
class TestInvokeComponentCommand(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to invoke a component command")
    def test_invoke_component_command(
        self, mocker, mock_digital_twin_operations, digital_twin_client
    ):
        ret_val = digital_twin_client.invoke_component_command(
            fake_digital_twin_id, fake_component_path, fake_component_name, fake_payload
        )
        assert mock_digital_twin_operations.invoke_component_command.call_count == 1
        assert mock_digital_twin_operations.invoke_component_command.call_args == mocker.call(
            fake_digital_twin_id, fake_component_path, fake_component_name, fake_payload, None, None
        )
        assert ret_val == mock_digital_twin_operations.invoke_component_command()


@pytest.mark.describe("DigitalTwinClient - .invoke_component_command() - optional parameters")
class TestInvokeComponentCommandWithOptionalParameters(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to invoke a component command")
    def test_invoke_component_command(
        self, mocker, mock_digital_twin_operations, digital_twin_client
    ):
        ret_val = digital_twin_client.invoke_component_command(
            fake_digital_twin_id,
            fake_component_path,
            fake_component_name,
            fake_payload,
            fake_conn_timeout,
            fake_resp_timeout,
        )
        assert mock_digital_twin_operations.invoke_component_command.call_count == 1
        assert mock_digital_twin_operations.invoke_component_command.call_args == mocker.call(
            fake_digital_twin_id,
            fake_component_path,
            fake_component_name,
            fake_payload,
            fake_conn_timeout,
            fake_resp_timeout,
        )
        assert ret_val == mock_digital_twin_operations.invoke_component_command()


@pytest.mark.describe("DigitalTwinClient - .invoke_command()")
class TestInvokeCommand(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to invoke a component command")
    def test_invoke_command(self, mocker, mock_digital_twin_operations, digital_twin_client):
        ret_val = digital_twin_client.invoke_command(
            fake_digital_twin_id, fake_component_name, fake_payload
        )
        assert mock_digital_twin_operations.invoke_root_level_command.call_count == 1
        assert mock_digital_twin_operations.invoke_root_level_command.call_args == mocker.call(
            fake_digital_twin_id, fake_component_name, fake_payload, None, None
        )
        assert ret_val == mock_digital_twin_operations.invoke_root_level_command()


@pytest.mark.describe("DigitalTwinClient - .invoke_command()")
class TestInvokeCommandWithOptionalParameters(object):
    @pytest.mark.it("Uses protocol layer DigitalTwin Client runtime to invoke a component command")
    def test_invoke_command(self, mocker, mock_digital_twin_operations, digital_twin_client):
        ret_val = digital_twin_client.invoke_command(
            fake_digital_twin_id,
            fake_component_name,
            fake_payload,
            fake_conn_timeout,
            fake_resp_timeout,
        )
        assert mock_digital_twin_operations.invoke_root_level_command.call_count == 1
        assert mock_digital_twin_operations.invoke_root_level_command.call_args == mocker.call(
            fake_digital_twin_id,
            fake_component_name,
            fake_payload,
            fake_conn_timeout,
            fake_resp_timeout,
        )
        assert ret_val == mock_digital_twin_operations.invoke_root_level_command()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import base64
import time
import hashlib
import hmac
from uuid import uuid4
import six.moves.urllib as urllib
from azure.core.credentials import AccessToken
import uamqp


default_sas_expiry = 3600


class IoTHubAmqpClientBase:
    def disconnect_sync(self):
        """
        Disconnect the Amqp client.
        """
        if self.amqp_client:
            self.amqp_client.close()
            self.amqp_client = None

    def send_message_to_device(self, device_id, message, app_props):
        """Send a message to the specified deivce.

        :param str device_id: The name (Id) of the device.
        :param str message: The message that is to be delivered to the device.
        :param dict app_props: Application and system properties for the message

        :raises: Exception if the Send command is not able to send the message
        """
        msg_content = message
        msg_props = uamqp.message.MessageProperties()
        msg_props.message_id = str(uuid4())
        msg_props.to = "/devices/{}/messages/devicebound".format(device_id)

        app_properties = {}

        # loop through all properties and pull out the custom
        # properties
        for prop_key, prop_value in app_props.items():
            if prop_key == "contentType":
                msg_props.content_type = prop_value
            elif prop_key == "contentEncoding":
                msg_props.content_encoding = prop_value
            elif prop_key == "correlationId":
                msg_props.correlation_id = prop_value
            elif prop_key == "expiryTimeUtc":
                msg_props.absolute_expiry_time = prop_value
            elif prop_key == "messageId":
                msg_props.message_id = prop_value
            else:
                app_properties[prop_key] = prop_value

        message = uamqp.Message(
            msg_content, properties=msg_props, application_properties=app_properties
        )
        self.amqp_client.queue_message(message)
        results = self.amqp_client.send_all_messages(close_on_done=False)
        if uamqp.constants.MessageState.SendFailed in results:
            raise Exception("C2D message send failure")


class IoTHubAmqpClientSharedAccessKeyAuth(IoTHubAmqpClientBase):
    def __init__(self, hostname, shared_access_key_name, shared_access_key, transport_type=uamqp.TransportType.Amqp):
        def get_token():
            expiry = int(time.time() + default_sas_expiry)
            sas = base64.b64decode(shared_access_key)
            string_to_sign = (hostname + "\n" + str(expiry)).encode("utf-8")
            signed_hmac_sha256 = hmac.HMAC(sas, string_to_sign, hashlib.sha256)
            signature = urllib.parse.quote(base64.b64encode(signed_hmac_sha256.digest()))
            return AccessToken(
                "SharedAccessSignature sr={}&sig={}&se={}&skn={}".format(
                    hostname, signature, expiry, shared_access_key_name
                ),
                expiry,
            )

        auth = uamqp.authentication.JWTTokenAuth(
            audience="https://" + hostname,
            uri="https://" + hostname,
            get_token=get_token,
            token_type=b"servicebus.windows.net:sastoken",
            transport_type=transport_type,
        )
        auth.update_token()
        self.amqp_client = uamqp.SendClient(
            target="amqps://" + hostname + "/messages/devicebound",
            auth=auth,
            keep_alive_interval=120,
            transport_type=transport_type,
        )


class IoTHubAmqpClientTokenAuth(IoTHubAmqpClientBase):
    def __init__(
        self, hostname, token_credential, token_scope="https://iothubs.azure.net/.default", transport_type=uamqp.TransportType.Amqp
    ):
        def get_token():
            result = token_credential.get_token(token_scope)
            return AccessToken("Bearer " + result.token, result.expires_on)

        auth = uamqp.authentication.JWTTokenAuth(
            audience=token_scope,
            uri="https://" + hostname,
            get_token=get_token,
            token_type=b"bearer",
            transport_type=transport_type
        )
        auth.update_token()
        target = "amqps://" + hostname + "/messages/devicebound"
        self.amqp_client = uamqp.SendClient(target=target, auth=auth, keep_alive_interval=120, transport_type=transport_type)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import pytest
from azure.iot.hub.iothub_configuration_manager import IoTHubConfigurationManager
from azure.iot.hub.auth import ConnectionStringAuthentication
from azure.iot.hub.protocol.iot_hub_gateway_service_ap_is import IotHubGatewayServiceAPIs

"""---Constants---"""

fake_shared_access_key = "Zm9vYmFy"
fake_shared_access_key_name = "alohomora"
fake_hostname = "beauxbatons.academy-net"
fake_device_id = "MyPensieve"
fake_etag = "taggedbyministryofmagic"
fake_configuration_id = "fake_configuration_id"


class fake_configuration_object:
    id = fake_configuration_id


fake_configuration = fake_configuration_object()
fake_max_count = 42
fake_configuration_queries = "fake_configuration_queries"
fake_configuration_content = "fake_configuration_content"


"""----Shared fixtures----"""


@pytest.fixture(scope="function", autouse=True)
def mock_configuration_operations(mocker):
    mock_configuration_operations_init = mocker.patch(
        "azure.iot.hub.protocol.iot_hub_gateway_service_ap_is.ConfigurationOperations"
    )
    return mock_configuration_operations_init.return_value


@pytest.fixture(scope="function")
def iothub_configuration_manager():
    connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
        hostname=fake_hostname,
        device_id=fake_device_id,
        skn=fake_shared_access_key_name,
        sk=fake_shared_access_key,
    )
    iothub_configuration_manager = IoTHubConfigurationManager.from_connection_string(
        connection_string
    )
    return iothub_configuration_manager


@pytest.mark.describe("IoTHubConfigurationManager - .from_connection_string()")
class TestFromConnectionString(object):
    @pytest.mark.parametrize(
        "connection_string",
        [
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    device_id=fake_device_id,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string with HostName, DeviceId, SharedAccessKeyName, and SharedAccessKey",
            ),
            pytest.param(
                "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                    hostname=fake_hostname,
                    skn=fake_shared_access_key_name,
                    sk=fake_shared_access_key,
                ),
                id="connection string without DeviceId",
            ),
            pytest.param(
                "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
                    hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
                ),
                id="connection string without SharedAccessKeyName",
            ),
        ],
    )
    @pytest.mark.it(
        "Creates an instance of ConnectionStringAuthentication and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_connection_string_auth(self, connection_string):
        client = IoTHubConfigurationManager.from_connection_string(
            connection_string=connection_string
        )

        assert repr(client.auth) == connection_string
        assert client.protocol.config.base_url == "https://" + client.auth["HostName"]
        assert client.protocol.config.credentials == client.auth

    @pytest.mark.it("Sets the auth and protocol attributes")
    def test_instantiates_auth_and_protocol_attributes(self, iothub_configuration_manager):
        assert isinstance(iothub_configuration_manager.auth, ConnectionStringAuthentication)
        assert isinstance(iothub_configuration_manager.protocol, IotHubGatewayServiceAPIs)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with an empty connection string"
    )
    def test_instantiates_with_empty_connection_string(self):
        with pytest.raises(ValueError):
            IoTHubConfigurationManager.from_connection_string("")

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without HostName"
    )
    def test_instantiates_with_connection_string_no_host_name(self):
        connection_string = (
            "DeviceId={device_id};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                device_id=fake_device_id, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        with pytest.raises(ValueError):
            IoTHubConfigurationManager.from_connection_string(connection_string)

    @pytest.mark.it("Instantiates with an connection string without DeviceId")
    def test_instantiates_with_connection_string_no_device_id(self):
        connection_string = (
            "HostName={hostname};SharedAccessKeyName={skn};SharedAccessKey={sk}".format(
                hostname=fake_hostname, skn=fake_shared_access_key_name, sk=fake_shared_access_key
            )
        )
        obj = IoTHubConfigurationManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubConfigurationManager)

    @pytest.mark.it("Instantiates with an connection string without SharedAccessKeyName")
    def test_instantiates_with_connection_string_no_shared_access_key_name(self):
        connection_string = "HostName={hostname};DeviceId={device_id};SharedAccessKey={sk}".format(
            hostname=fake_hostname, device_id=fake_device_id, sk=fake_shared_access_key
        )
        obj = IoTHubConfigurationManager.from_connection_string(connection_string)
        assert isinstance(obj, IoTHubConfigurationManager)

    @pytest.mark.it(
        "Raises a ValueError exception when instantiated with a connection string without SharedAccessKey"
    )
    def test_instantiates_with_connection_string_no_shared_access_key(self):
        connection_string = (
            "HostName={hostname};DeviceId={device_id};SharedAccessKeyName={skn}".format(
                hostname=fake_hostname, device_id=fake_device_id, skn=fake_shared_access_key_name
            )
        )
        with pytest.raises(ValueError):
            IoTHubConfigurationManager.from_connection_string(connection_string)


@pytest.mark.describe("IoTHubConfigurationManager - .from_token_credential()")
class TestFromTokenCredential(object):
    @pytest.mark.it(
        "Creates an instance of AzureIdentityCredentialAdapter and passes it to IotHubGatewayServiceAPIs constructor"
    )
    def test_token_credential_auth(self, mocker):
        mock_azure_identity_TokenCredential = mocker.MagicMock()

        client = IoTHubConfigurationManager.from_token_credential(
            fake_hostname, mock_azure_identity_TokenCredential
        )

        assert client.auth._policy._credential == mock_azure_identity_TokenCredential
        assert client.protocol.config.base_url == "https://" + fake_hostname
        assert client.protocol.config.credentials == client.auth


@pytest.mark.describe("IoTHubConfigurationManager - .get_configuration()")
class TestGetConfiguration(object):
    @pytest.mark.it("Gets configuration")
    def test_get(self, mocker, mock_configuration_operations, iothub_configuration_manager):
        iothub_configuration_manager.get_configuration(fake_configuration_id)

        assert mock_configuration_operations.get.call_count == 1
        assert mock_configuration_operations.get.call_args == mocker.call(fake_configuration_id)


@pytest.mark.describe("IoTHubConfigurationManager - .create_configuration()")
class TestCreateConfiguration(object):
    @pytest.mark.it("Creates configuration")
    def test_create_configuration(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.create_configuration(fake_configuration)

        assert mock_configuration_operations.create_or_update.call_count == 1
        assert mock_configuration_operations.create_or_update.call_args == mocker.call(
            fake_configuration_id, fake_configuration
        )


@pytest.mark.describe("IoTHubConfigurationManager - .update_configuration()")
class TestUpdateConfiguration(object):
    @pytest.mark.it("Updates configuration")
    def test_update_configuration(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.update_configuration(fake_configuration, fake_etag)

        assert mock_configuration_operations.create_or_update.call_count == 1
        assert mock_configuration_operations.create_or_update.call_args == mocker.call(
            fake_configuration_id, fake_configuration, fake_etag
        )


@pytest.mark.describe("IoTHubConfigurationManager - .delete_configuration()")
class TestDeleteConfiguration(object):
    @pytest.mark.it("Deletes configuration")
    def test_delete_configuration(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.delete_configuration(fake_configuration_id)

        assert mock_configuration_operations.delete.call_count == 1
        assert mock_configuration_operations.delete.call_args == mocker.call(
            fake_configuration_id, "*"
        )

    @pytest.mark.it("Deletes configuration with an etag")
    def test_delete_configuration_with_etag(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.delete_configuration(
            configuration_id=fake_configuration_id, etag=fake_etag
        )

        assert mock_configuration_operations.delete.call_count == 1
        assert mock_configuration_operations.delete.call_args == mocker.call(
            fake_configuration_id, fake_etag
        )


@pytest.mark.describe("IoTHubConfigurationManager - .get_configurations()")
class TestGetConfigurations(object):
    @pytest.mark.it("Get configurations")
    def test_get_configurations(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.get_configurations(fake_max_count)

        assert mock_configuration_operations.get_configurations.call_count == 1
        assert mock_configuration_operations.get_configurations.call_args == mocker.call(
            fake_max_count
        )


@pytest.mark.describe("IoTHubConfigurationManager - .test_configuration_queries()")
class TestTestConfigurationQueries(object):
    @pytest.mark.it("Test test_configuration_queries")
    def test_test_configuration_queries(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.test_configuration_queries(fake_configuration_queries)
        assert mock_configuration_operations.test_queries.call_count == 1
        assert mock_configuration_operations.test_queries.call_args == mocker.call(
            fake_configuration_queries
        )


@pytest.mark.describe("IoTHubConfigurationManager - .apply_configuration_on_edge_device()")
class TestApplyConfigurationOnEdgeDevice(object):
    @pytest.mark.it("Test apply configuration on edge device")
    def test_apply_configuration_on_edge_device(
        self, mocker, mock_configuration_operations, iothub_configuration_manager
    ):
        iothub_configuration_manager.apply_configuration_on_edge_device(
            fake_device_id, fake_configuration_content
        )
        assert mock_configuration_operations.apply_on_edge_device.call_count == 1
        assert mock_configuration_operations.apply_on_edge_device.call_args == mocker.call(
            fake_device_id, fake_configuration_content
        )
